﻿using System;
using System.ComponentModel;
using System.IO;
using System.Windows;
using System.Windows.Input;
using Byond.Messages;
using Byond.Persistence;
using Byond.Shell;
using Byond.Standalone.MidiDotNetStuff;
using Byond.Standalone.PortAudioStuff;
using PortAudioSharp;
using Utils;
using Utils.DialogManagement;
using Utils.Messaging;

namespace Byond.Standalone
{
	/// <summary>
	/// Interaction logic for MainWindow.xaml
	/// </summary>
	public partial class MainWindow : INotifyPropertyChanged
	{
		private readonly string _sessionPersistenceFileName;

		private readonly SettingsManager _settingsManager;
		private readonly PortAudioEngine _audioEngine;
		private readonly DotNetMidi _dotNetMidi;
		private readonly PersistenceManager _persistenceManager;
		private readonly IDialogManager _dialogManager;

		internal MainWindow(
			PortAudioEngine audioEngine,
			ByondControl samplumControl,
			PersistenceManager persistenceManager,
			Env env)
		{
			//AvalonEditor.UseRichTextBox = true;

			_sessionPersistenceFileName = env.RootPath + "\\LastSession.dat";
			_audioEngine = audioEngine;
			_persistenceManager = persistenceManager;

			MinimizeWindowStateCommand = new RelayCommand(() => WindowState = WindowState.Minimized);
			RestoreWindowStateCommand = new RelayCommand(() => WindowState = WindowState.Normal);
			MaximizeWindowStateCommand = new RelayCommand(() => WindowState = WindowState.Maximized);
			CloseWindowCommand = new RelayCommand(Close);
			StateChanged += (s, e) =>
				{
					OnPropertyChanged("CanRestoreWindowState");
					OnPropertyChanged("CanMaximizeWindowState");
				};

			DataContext = this;
			InitializeComponent();

			byondControlPlaceholder.Content = samplumControl;
			Loaded += OnLoaded;
		}

		public event MessageEventHandler SendMessage;

		protected override void OnClosing(CancelEventArgs e)
		{
			base.OnClosing(e);
			SendMessage(new SaveAllEnsemblesRequest());
			Persist();
		}

		private void OnLoaded(object sender, RoutedEventArgs routedEventArgs)
		{
			Loaded -= OnLoaded;

			// TODO: Explizitere Initialisierungs-Workflow machen.
			// Z.B. wegen: AudioEngine wird gestartet, obwohl noch kein UManager da ist.
			// In VST ist das anders gelöst. Vereinheitlichen?
			SendMessage(new InitializationStage1Message());
			SendMessage(new InitializationStage2Message());
		}

		public void HandleMessage(InitializationStage2Message message)
		{
			try
			{
				var midiSettings = _settingsManager.TryGetSettings<MidiSettings>();
				if (midiSettings != null)
				{
					var inputDevice = midiSettings.GetInputDevice();
					if (inputDevice != null)
						_dotNetMidi.SetInputDevice(inputDevice);

					var outputDevice = midiSettings.GetOutputDevice();
					if (outputDevice != null)
						_dotNetMidi.SetOutputDevice(outputDevice);
				}
				else
					_dotNetMidi.UseDefaults();
			}
			catch (Exception e)
			{
				_dialogManager
					.CreateMessageDialog(e.Message, "Could not initialize or restore midi engine.", DialogMode.Ok)
					.Show();
			}

			// Initialize Port Audio and read from last settings
			try
			{
				PortAudioExtensions.Initialize();

				var apiInfo = PortAudioExtensions.GetDefaultApi();
				var inputDevice = PortAudio.Pa_GetDeviceInfo(apiInfo.defaultInputDevice);
				var outputDevice = PortAudio.Pa_GetDeviceInfo(apiInfo.defaultOutputDevice);
				var sampleRate = (int)outputDevice.defaultSampleRate;
				var bufferSize = _audioEngine.BufferSize;

				var audioSettings = _settingsManager.TryGetSettings<AudioSettings>();
				if (audioSettings != null)
				{
					try
					{
						inputDevice = audioSettings.GetInputDevice();
						outputDevice = audioSettings.GetOutputDevice();
						sampleRate = audioSettings.SampleRate;
						bufferSize = audioSettings.BufferSize;
					}
					catch (Exception e)
					{
						_dialogManager
							.CreateMessageDialog(e.Message, "Could not restore audio settings. Using defaults.", DialogMode.Ok)
							.Show();
					}
				}

				_audioEngine.ReStart(sampleRate, bufferSize, inputDevice, outputDevice);
			}
			catch (Exception e)
			{
				_dialogManager
					.CreateMessageDialog(e.Message, "Could not initialize or restore audio engine.", DialogMode.Ok)
					.Show();
			}

			// Restore last session
			try
			{
				if (File.Exists(_sessionPersistenceFileName))
				{
					var json = File.ReadAllText(_sessionPersistenceFileName);
					_persistenceManager.RestoreEnsemble(json);
				}
			}
			catch (Exception e)
			{
				_dialogManager
					.CreateMessageDialog(e.Message, "Could not restore last session.", DialogMode.Ok)
					.Show();
			}
		}

		private void Persist()
		{
			if (!_persistenceManager.ShouldPersistEnsemble)
				return;

			var json = _persistenceManager.PersistEnsembleToJson();
			File.WriteAllText(_sessionPersistenceFileName, json);
		}

		#region Window Minimize / Maximize / Close Buttons

		public ICommand MinimizeWindowStateCommand { get; private set; }
		public ICommand RestoreWindowStateCommand { get; private set; }
		public ICommand MaximizeWindowStateCommand { get; private set; }
		public ICommand CloseWindowCommand { get; private set; }

		public bool CanRestoreWindowState { get { return WindowState == WindowState.Maximized; } }
		public bool CanMaximizeWindowState { get { return WindowState != WindowState.Maximized; } }

		protected override void OnMouseLeftButtonDown(MouseButtonEventArgs e)
		{
			base.OnMouseLeftButtonDown(e);

			if (e.ClickCount != 1)
				return;
			DragMove();
			e.Handled = true;
		}

		#endregion

		public event PropertyChangedEventHandler PropertyChanged;
		private void OnPropertyChanged(string propertyName)
		{
			var handler = PropertyChanged;
			if (handler != null)
				handler(this, new PropertyChangedEventArgs(propertyName));
		}
	}
}
