// Copyright 2013 Ronald Schlenker, www.technewlogic.de.
// 
// This file is part of Byond.
// 
// GraphIT is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 2 of the License, or
// (at your option) any later version.
// 
// Foobar is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
using System;
using System.Diagnostics;
using System.Linq;
using Byond.HostObjects;
using Byond.Logging;
using Byond.Messages;
using Byond.Recording;
using Utils.Messaging;

namespace Byond.AudioMidi
{
	public abstract class AudioEngine : IAudioInfo
	{
		public event MessageEventHandler SendMessage;

		private readonly RuntimeScope _runtimeScope;
		private readonly Scheduler _scheduler;
		private readonly AudioRecorder _audioRecorder;
		private readonly DiagnosisManager _diagnosisManager;
		private readonly ILogger _logger;

		private const int UiBindTimesPerSec = 100;
		private int _uiBindStepWidth;
		private int _currentUiBindingCycle;

		private const int UiSmoothTimesPerSec = 1000;
		private int _uiSmoothStepWidth;
		private int _currentUiSmoothingCycle;

		private const int SchedulerUpdateTimesPerSec = 500;
		private int _schedulerStepWidth;
		private int _currentSchedulerCycle;

		private const int EventProcessingTimesPerSec = 200;
		private int _eventProcessingStepWidth;
		private int _currentEventProcessingCycle;

		private int _last96Note;

		protected int CurrentCycle { get; private set; }

		public abstract double SongTempo { get; }
		public abstract double SampleRate { get; }
		public abstract double SamplePosition { get; }
		public abstract bool IsTransportPlaying { get; }
		protected abstract int SampleCountInCycle { get; }

		public bool IsRunning { get; private set; }

		//protected abstract void OnVoicesProcessing();
		protected abstract void OnVoicesProcessed();
		protected abstract double[] GetOutputChannelSamples(int channel);

		public void HandleMessage(EnsembleRunningMessage message)
		{
			IsRunning = true;
		}

		public void HandleMessage(EnsembleStoppedMessage message)
		{
			IsRunning = false;
		}

		public void ProcessAudio()
		{
			if (!IsRunning)
				return;

			try
			{
				_diagnosisManager.StartCpuMeasure();

				// Get Time Infos
				var songTempo = SongTempo;
				var sampleRate = SampleRate;
				var originalSamplePos = SamplePosition;
				var samplePos = originalSamplePos;

				var runtimeScopeContext = _runtimeScope.CurrentContext;
				var patchInfo = runtimeScopeContext.PatchInfo;
				var patchManager = runtimeScopeContext.PatchManager;
				var voiceManager = runtimeScopeContext.VoiceManager;
				var keyboardCallbackManager = runtimeScopeContext.KeyboardCallbackManager;
				var midiCallbackManager = runtimeScopeContext.MidiCallbackManager;
				var uiManager = runtimeScopeContext.UiManager;
				var currentPatch = patchManager.RootPatch;
				if (currentPatch == Patch.Empty)
					return;

				// Assign values to Patch
				patchInfo.SongTempo = songTempo;
				patchInfo.SampleRate = sampleRate;
				var wasPlaying = patchInfo.IsPlaying;
				var isPlaying = IsTransportPlaying;
				patchInfo.IsPlaying = isPlaying;
				patchInfo.PlayingStateChanged = isPlaying != wasPlaying;

				// Reset UI / Scheduler Step Width
				_uiBindStepWidth = (int)(sampleRate / UiBindTimesPerSec);
				_uiSmoothStepWidth = (int)(sampleRate / UiSmoothTimesPerSec);
				_schedulerStepWidth = (int)(sampleRate / SchedulerUpdateTimesPerSec);
				_eventProcessingStepWidth = (int)(sampleRate / EventProcessingTimesPerSec);

				CurrentCycle = 0;
				var sampleCountInCycle = SampleCountInCycle;
				for (; CurrentCycle < sampleCountInCycle; CurrentCycle++)
				{
					// IMP: Warum geht das hier nicht in Cubase (mit samplePos)? Note96 bekommt dann immer ganz komische Werte
					var currentSamplePos = (isPlaying ? samplePos : originalSamplePos);
					//var currentSamplePos = samplePos;

					// Tempo
					var ppq = currentSamplePos * songTempo / (60 * sampleRate);
					patchInfo.Ppq = ppq;
					patchInfo.NextPpq = (currentSamplePos + 1) * songTempo / (60 * sampleRate);
					patchInfo.SamplePos = currentSamplePos;

					// 96 Note
					var leadingPpq = (int)ppq;
					// Ppq: 1/4 Note entspricht 1, d.h.: 4 entspricht einem Takt
					var bar = leadingPpq / 4;
					var noteInBar = ppq - bar * 4;
					var position = new Position(bar, (int)(noteInBar * 24));
					patchInfo.Pos = position;

					if (voiceManager.EnableVoiceProcessing)
					{
						//OnVoicesProcessing();
						voiceManager.ProcessVoices();
						if (currentPatch.ConnectCallback != null)
							currentPatch.ConnectCallback();
						OnVoicesProcessed();
					}

					samplePos++;

					// War vorher vor dem voiceProcessing Block. Ist das ok, wenn es hier gemacht wird?
					if (_last96Note != position.Note96)
						currentPatch.RaiseNote96Event();
					_last96Note = position.Note96;

					_currentUiBindingCycle++;
					if (_currentUiBindingCycle == _uiBindStepWidth)
					{
						_currentUiBindingCycle = 0;
						uiManager.RefreshBindings();
						uiManager.InvokeClickActions();
					}

					_currentUiSmoothingCycle++;
					if (_currentUiSmoothingCycle == _uiSmoothStepWidth)
					{
						_currentUiSmoothingCycle = 0;
						uiManager.RefreshSmoothings();
					}

					_currentSchedulerCycle++;
					if (_currentSchedulerCycle == _schedulerStepWidth)
					{
						_currentSchedulerCycle = 0;
						_scheduler.ProcessScheduling();
					}

					_currentEventProcessingCycle++;
					if (_currentEventProcessingCycle == _eventProcessingStepWidth)
					{
						_currentEventProcessingCycle = 0;
						keyboardCallbackManager.Raise();
						midiCallbackManager.Raise();
					}
				}

				if (_audioRecorder.IsWriting)
				{
					var channelData = Enumerable.Range(0, 2).Select(GetOutputChannelSamples).ToArray();
					_audioRecorder.Write(channelData);
				}

				_diagnosisManager.StopCpuMeasure(sampleCountInCycle, sampleRate);
				_diagnosisManager.VoicesCount = voiceManager.VoicesCount;
			}
			catch (Exception e)
			{
				_logger.Log("Error in Voice Processing / Connect: " + e);
				IsRunning = false;
				SendMessage(new StopEnsembleRequest(e));
			}
		}
	}
}