// Copyright 2013 Ronald Schlenker, www.technewlogic.de.
// 
// This file is part of Byond.
// 
// GraphIT is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 2 of the License, or
// (at your option) any later version.
// 
// Foobar is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
using System;
using System.CodeDom.Compiler;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Runtime.CompilerServices;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Markup;
using System.Xml.Linq;
using Byond.EnsembleLibrary;
using Byond.HostObjects;
using Microsoft.CSharp;

namespace Byond.Compilation
{
	sealed class PatchCompiler
	{
		private readonly Library _library;
		private readonly PatchBuilder _patchBuilder;
		private readonly Func<Ensemble, IEnumerable<Ensemble>, PatchClassPreparer> _patchClassPreparerFactory;

		public PatchCompiler()
		{
			AppDomain.CurrentDomain.AssemblyResolve += (sender, args) =>
				AppDomain.CurrentDomain.GetAssemblies().FirstOrDefault(it => it.FullName == args.Name);
		}

		private bool _patchGenerated;

		public Patch GeneratePatch(
			Ensemble ensemble,
			IEnumerable<Ensemble> allEnsembles)
		{
			if (_patchGenerated)
				throw new Exception("PatchCompiler cannot be used twice.");
			_patchGenerated = true;

			var pcp = _patchClassPreparerFactory(ensemble, allEnsembles);
			// GeneratePatch wird hier nicht vom Ctor aufgerufen, da es dann zu einer
			// DependencyResolutionException kommen wrde und das den Dialog erschwerewn wrde.
			// TODO: Exception-Typ anlegen, der generisch gefangen und angezeigt wird.
			pcp.GeneratePatch();

			var loadedAssemblies = LoadImportAssemblies(pcp);
			var assembly = CompileAssembly(pcp);

			var patchTypes = assembly
				.GetTypes()
				.Where(it => typeof(Patch).IsAssignableFrom(it));

			var patchType = patchTypes.FirstOrDefault(it => it.Name == pcp.RootPatchName);
			if (patchType == null)
				throw new Exception("No Root Patch Found.");

			// Pre-Jit all the assemblies
			new[] {assembly}.Concat(loadedAssemblies).ForEach(PreJitAssembly);

			_patchBuilder.RegisterAssemblyTypes(assembly);
			return _patchBuilder.ResolvePatch(patchType, null, null, true);
		}

		private IEnumerable<Assembly> LoadImportAssemblies(PatchClassPreparer pcp)
		{
			return pcp.Imports
				.Select(it => Assembly.LoadFrom(
					Path.Combine(_library.DependenciesDirectory, it + ".dll")))
				.ToList();
		}

		private Assembly CompileAssembly(PatchClassPreparer pcp)
		{
			var provider = new CSharpCodeProvider(
				new Dictionary<string, string> { { "CompilerVersion", "v4.0" } });
			var cp = new CompilerParameters();

			var ver = string.Format("{0}.{1}.{2}", Environment.Version.Major, Environment.Version.MajorRevision, Environment.Version.Build);
			var exWpfDir = string.Format(@"C:\WINDOWS\Microsoft.NET\Framework\v{0}\WPF", ver);
			var exDir = string.Format(@"C:\WINDOWS\Microsoft.NET\Framework\v{0}", ver);

			new[]
				{
					typeof(Action<>).Assembly.Location,
					//typeof(object).Assembly.Location, // mscorelib
					typeof(Enumerable).Assembly.Location, // System.Core
					//typeof(FrameworkElement).Assembly.Location, // PresentationFramework
					typeof(UIElement).Assembly.Location, // PresentationCore
					typeof(DependencyObject).Assembly.Location, // WindowsBase
					//typeof(XamlReader).Assembly.Location, // System.Xaml
				    typeof(Debug).Assembly.Location, // System
				    GetType().Assembly.Location
				}
				.Concat(pcp.Imports
					.Select(it => Path.Combine(_library.DependenciesDirectory, it + ".dll"))
					.ToList())
				.ForEach(it => cp.ReferencedAssemblies.Add(it));

			cp.GenerateExecutable = false;
			cp.GenerateInMemory = true;
			cp.TreatWarningsAsErrors = false;
			cp.CompilerOptions = string.Format("/lib:{0}", exWpfDir);
			cp.CompilerOptions = string.Format("/lib:{0}", exDir);

			//if (!Debugger.IsAttached)
			//{
			cp.CompilerOptions = "/optimize+";
			cp.IncludeDebugInformation = false;
			//}
			//else
			//    cp.IncludeDebugInformation = true;

			var result = provider.CompileAssemblyFromFile(
				cp,
				pcp.SourceCodeFiles.Select(it => it.TempFileName).ToArray());
			if (result.Errors.HasErrors)
			{
				var errorMessage = "\r\n------------ COMPILATION ERRORS ------------";
				foreach (CompilerError error in result.Errors)
				{
					var sourceFile = pcp.SourceCodeFiles
						.First(it => Path.GetFileName(it.TempFileName) == Path.GetFileName(error.FileName));
					// TODO: Line in Error Text mit Offsets und Mappings machen
					errorMessage += "\r\n"
						+ "'" +
						sourceFile.ClassName + "'"
						+ " in  LINE: " + sourceFile.MapLine(error.Line)
						+ "  -  " + error.ErrorText;
				}
				errorMessage += "\r\n------------ END ERRORS ------------\r\n";

				foreach (var sourceCodeFile in pcp.SourceCodeFiles)
				{
					errorMessage += File
						.ReadAllLines(sourceCodeFile.TempFileName)
						.SkipWhile(string.IsNullOrWhiteSpace)
						.Select((it, i) => (i + 1) + "  " + it)
						.Aggregate(
							"\r\n" + sourceCodeFile.ClassName + ":",
							(a, b) => a + "\r\n" + b);
				}

				throw new Exception(errorMessage);
			}

			return result.CompiledAssembly;
		}

		private static void PreJitAssembly(Assembly assembly)
		{
			assembly
				.GetTypes()
				.SelectMany(it => it.GetMethods(
					BindingFlags.DeclaredOnly
					| BindingFlags.NonPublic
					| BindingFlags.Public
					| BindingFlags.Instance
					| BindingFlags.Static))
				.Where(it => !it.IsAbstract
					&& !it.ContainsGenericParameters
					&& !it.IsGenericMethod
					&& !it.IsGenericMethodDefinition
					&& ((it.Attributes & MethodAttributes.PinvokeImpl) <= 0))
				.ForEach(it => RuntimeHelpers.PrepareMethod(it.MethodHandle));
		}

		public static Control GenerateUi(string xaml)
		{
			xaml = @"<wpf:UserControl 
	xmlns:wpf='http://schemas.microsoft.com/winfx/2006/xaml/presentation'
	xmlns:x='http://schemas.microsoft.com/winfx/2006/xaml'
	xmlns:uiRuntime='clr-namespace:Byond.UIRuntime;assembly=Byond.Core'
	xmlns='clr-namespace:Byond.PatchUiControls;assembly=Byond.Core'>"
				+ xaml + "\r\n</wpf:UserControl>";
			var xml = XDocument.Parse(xaml);
			xml.Descendants().Attributes()
				.Where(it => it.Value.StartsWith("$"))
				.ForEach(it => it.Value = "{uiRuntime:UiBindingMarkupExtension " + it.Value.Substring(1) + "}");
			var control = XamlReader.Parse(xml.ToString()) as Control;
			return control;
		}
	}
}
