// Copyright 2013 Ronald Schlenker, www.technewlogic.de.
// 
// This file is part of Byond.
// 
// GraphIT is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 2 of the License, or
// (at your option) any later version.
// 
// Foobar is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using Byond.Messages;
using Utils.DialogManagement;
using Utils.Messaging;

namespace Byond.EnsembleLibrary
{
	sealed class Library
	{
		public void Ctor()
		{
			if (!Directory.Exists(_env.RootPath))
				_needsFactoryInitialize = true;

			LibraryDirectory = _env.PatchPath;
			if (!Directory.Exists(LibraryDirectory))
				Directory.CreateDirectory(LibraryDirectory);

			DependenciesDirectory = _env.AssemblyPath;
			if (!Directory.Exists(DependenciesDirectory))
				Directory.CreateDirectory(DependenciesDirectory);
		}

		private readonly Env _env;

		private bool _needsFactoryInitialize;
		private readonly IDialogManager _dialogManager;

		public event MessageEventHandler SendMessage;

		public const string PatchFileExtension = ".byond";

		public string LibraryDirectory { get; private set; }
		public string DependenciesDirectory { get; private set; }
		public LibraryFolder RootFolder { get; private set; }

		public void HandleMessage(EnsemblesUpdatedMessage message)
		{
			Refresh();
		}

		public void HandleMessage(InitializationStage1Message message)
		{
			if (!_needsFactoryInitialize)
			{
				Refresh();
				return;
			}

			var dialog = _dialogManager.CreateMessageDialog(
				"Do you want to create a default library in your Documents folder?",
				"No library was found",
				DialogMode.YesNo);
			dialog.Yes = CreateFactoryLibrary;
			dialog.Show();
		}

		public LibraryFolder TryGetFolder(string folderPath)
		{
			return RootFolder.MeAndDescendantFolders.FirstOrDefault(it => it.Path == folderPath)
				?? RootFolder;
		}

		private void CreateFactoryLibrary()
		{
			var sourcePath = Path.GetDirectoryName(GetType().Assembly.Location) + "\\FactoryLibrary";
			var destinationPath = _env.RootPath;

			CopyFolder(sourcePath, destinationPath);

			Refresh();
		}

		private static void CopyFolder(string sourceFolder, string destFolder)
		{
			if (!Directory.Exists(destFolder))
				Directory.CreateDirectory(destFolder);
			
			var files = Directory.GetFiles(sourceFolder);
			foreach (var file in files)
			{
				var name = Path.GetFileName(file);
				var dest = Path.Combine(destFolder, name);
				File.Copy(file, dest);
			}
			
			var folders = Directory.GetDirectories(sourceFolder);
			foreach (var folder in folders)
			{
				var name = Path.GetFileName(folder);
				var dest = Path.Combine(destFolder, name);
				CopyFolder(folder, dest);
			}
		}

		private void Refresh()
		{
			var di = new DirectoryInfo(LibraryDirectory);
			RootFolder = new LibraryFolder(di.FullName, GetItems(di), GetFolders(di));
			SendMessage(new LibraryUpdatedMessage());
		}

		private IEnumerable<LibraryFolder> GetFolders(DirectoryInfo di)
		{
			return di
				.GetDirectories("*", SearchOption.TopDirectoryOnly)
				.Select(CreateLibraryFolder)
				.ToList();
		}

		private IEnumerable<LibraryItem> GetItems(DirectoryInfo di)
		{
			return di.GetFiles("*" + PatchFileExtension).Select(CreateLibraryItem).ToList();
		}

		private LibraryItem CreateLibraryItem(FileInfo fileInfo)
		{
			return new LibraryItem(fileInfo.FullName);
		}

		private LibraryFolder CreateLibraryFolder(DirectoryInfo directoryInfo)
		{
			return new LibraryFolder(directoryInfo.FullName, GetItems(directoryInfo), GetFolders(directoryInfo));
		}
	}
}
