// Copyright 2013 Ronald Schlenker, www.technewlogic.de.
// 
// This file is part of Byond.
// 
// GraphIT is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 2 of the License, or
// (at your option) any later version.
// 
// Foobar is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.IO;
using System.Linq;
using Mvvm;
using Byond.Messages;
using Utils;
using Utils.Messaging;

namespace Byond.EnsembleLibrary.ViewModels
{
	public class BrowserViewModel : ViewModelBase
	{
		private readonly Library _library;
		private readonly EnsembleManager _ensembleManager;
		private readonly Func<Ensemble, ItemViewModel> _itemFactory;

		public event MessageEventHandler SendMessage;

		private FolderViewModel _currentFolder;

		private readonly List<ItemViewModel> _ensembleViewModels
			= new List<ItemViewModel>();

		private readonly ObservableCollection<ItemViewModel> _items = new ObservableCollection<ItemViewModel>();
		public IEnumerable<ItemViewModel> Items { get { return _items; } }

		private readonly ObservableCollection<FolderViewModel> _folders = new ObservableCollection<FolderViewModel>();
		public IEnumerable<FolderViewModel> Folders { get { return _folders; } }

		private readonly ObservableCollection<BreadcrumpViewModel> _breadcrump = new ObservableCollection<BreadcrumpViewModel>();
		public IEnumerable<BreadcrumpViewModel> Breadcrump { get { return _breadcrump; } }

		public void HandleMessage(LibraryUpdatedMessage message)
		{
			RefreshBrowser(
				_currentFolder == null
					? string.Empty
					: _currentFolder.Path);
		}

		public void HandleMessage(EnsembleRunningMessage message)
		{
			Save();
		}

		public void HandleMessage(LocateCurrentPatchRequest message)
		{
			TryGotoPatch(_ensembleManager.EditingEnsemble);
		}

		public void HandleMessage(EnsembleActivatedMessage message)
		{
			TryGotoPatch(_ensembleManager.EditingEnsemble);
		}

		private void TryGotoPatch(Ensemble patch)
		{
			if (patch == null)
				return;

			var path = Path.GetDirectoryName(patch.FileName);
			if (Directory.Exists(path))
				ShowFolder(path);
		}

		public void New()
		{
			SendMessage(new NewEnsembleRequest());
		}

		public void Save()
		{
			SendMessage(new SaveAllEnsemblesRequest());
		}

		public void SaveAs()
		{
			SendMessage(new SaveEnsembleAsRequest());
		}

		public void ShowActivePatch()
		{
			TryGotoPatch(_ensembleManager.ActiveEnsemble);
		}

		private void RefreshBrowser(string folderPath)
		{
			// Find Folder
			var libraryFolder = string.IsNullOrWhiteSpace(folderPath)
				? _library.RootFolder
				: _library.TryGetFolder(folderPath);

			_ensembleViewModels.Clear();
			_currentFolder = new FolderViewModel(
				libraryFolder,
				ToViewModel(libraryFolder.Items),
				ToViewModel(libraryFolder.Folders),
				ShowFolder);

			_items.Clear();
			_currentFolder.Items.ForEach(it => _items.Add(it));

			_folders.Clear();
			_currentFolder.Folders.ForEach(it => _folders.Add(it));
		}

		private void ShowFolder(string folder)
		{
			RefreshBrowser(folder);

			_breadcrump.Clear();

			//var currentElement = folder.Parent;
			//while (currentElement != null)
			//{
			//    _breadcrump.Insert(0, _breadcrumpFactory(currentElement));
			//    currentElement = currentElement.Parent;
			//}

			// wir wollen nur noch 1 einziges Element in der Breadcrump haben
			var libraryFolder = _library.TryGetFolder(folder);
			var currentElement = libraryFolder;
			if (currentElement != _library.RootFolder)
				_breadcrump.Add(
					new BreadcrumpViewModel(libraryFolder.Parent, "..", ShowFolder));
		}

		// TODO: Diese Methode muss beim beenden der Anwendung aufgerufen werden
		// TODO: hnlich wie bei SaveAs Dialog
		//private bool DiscardOfEnsembleAllowed()
		//{
		//    if (ActiveEnsembleViewModel == null || !_uow.IsDirty(ActiveEnsembleViewModel.Ensemble))
		//        return true;
		//    var res = MessageBox.Show(
		//        "Save this ensemble before making a new one?",
		//        "Attention",
		//        MessageBoxButton.YesNoCancel);
		//    switch (res)
		//    {
		//        case MessageBoxResult.Yes:
		//            SaveEnsembles();
		//            return true;
		//        case MessageBoxResult.No:
		//            return true;
		//        default:
		//            return false;
		//    }
		//}

		private IEnumerable<ItemViewModel> ToViewModel(IEnumerable<LibraryItem> items)
		{
			var viewModels = items
				.Select(it => _ensembleManager.Ensembles
					.FirstOrDefault(it2 => it2.FileName.Equals(it.Path, StringComparison.InvariantCultureIgnoreCase)))
				.Where(it => it != null)
				.Select(it => _itemFactory(it))
				.ToList();
			_ensembleViewModels.AddRange(viewModels);
			return viewModels;
		}

		private IEnumerable<FolderViewModel> ToViewModel(IEnumerable<LibraryFolder> folders)
		{
			return folders
				.Select(it => new FolderViewModel(
					it,
					ToViewModel(it.Items),
					ToViewModel(it.Folders),
					ShowFolder))
				.ToList();
		}
	}
}
