﻿// Copyright 2013 Ronald Schlenker, www.technewlogic.de.
// 
// This file is part of Byond.
// 
// GraphIT is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 2 of the License, or
// (at your option) any later version.
// 
// Foobar is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
using System;
using System.Reflection;
using System.Windows.Threading;
using Autofac.Core;
using Byond.HostObjects;
using Byond.Logging;
using Byond.Messages;
using Byond.Persistence;
using Utils.DialogManagement;
using Utils.Messaging;

namespace Byond
{
	class EnsembleRunner
	{
		private readonly IDialogManager _dialogManager;
		private readonly RuntimeScope _runtimeScope;
		private readonly EnsembleManager _ensembleManager;
		private readonly PersistenceManager _persistenceManager;
		private readonly Lazy<ILogger> _logger;
		private readonly Lazy<Dispatcher> _dispatcher;

		public event MessageEventHandler SendMessage;

		public void HandleMessage(RunActiveEnsembleRequest request)
		{
			var run = new Action(() => CreatePatch(_ensembleManager.ActiveEnsemble));
			var update = new Action(() =>
				{
					_persistenceManager.MarkForUpdate();
					CreatePatch(_ensembleManager.ActiveEnsemble);

				});

			var rootPatch = _runtimeScope.CurrentContext.PatchManager.RootPatch;
			if (rootPatch != Patch.Empty)
			{
				if (_ensembleManager.ActiveEnsemble.FileName.GetClassName() == rootPatch.GetType().Name)
				{
					var dialog = _dialogManager.CreateMessageDialog(
						"The state of the current patch will be overwritten.",
						"Stop current patch and replace it?",
						DialogMode.YesNoCancel);
					dialog.YesText = "Update";
					dialog.NoText = "Reset";
					dialog.Yes = update;
					dialog.No = run;
					dialog.Show();
				}
				else
				{
					var dialog = _dialogManager.CreateMessageDialog(
						"The state of the current patch will be overwritten.",
						"Stop current patch and replace it?",
						DialogMode.YesNo);
					dialog.Yes = run;
					dialog.Show();
				}
			}
			else
				run();
		}

		public void HandleMessage(StopEnsembleRequest request)
		{
			if (_dispatcher.Value.CheckAccess())
				_runtimeScope.Reset();
			else
				_dispatcher.Value.Invoke(_runtimeScope.Reset);

			if (request.Exception != null)
			{
				var dialog = _dialogManager
					.CreateMessageDialog(
						request.Exception.Message,
						"An Error occured in the patch code",
						DialogMode.Ok);

				dialog.Ok = () => SendMessage(new EnsembleStoppedMessage());
				dialog.Show();
			}
			else
				SendMessage(new EnsembleStoppedMessage());

			_logger.Value.Log("Ensemble stopped.");
		}

		private void CreatePatch(Ensemble ensemble)
		{
			// zuerst prüfen, dann richtiges Patch erstellen
			_runtimeScope.PrepareFutureContext();
			var exception = _runtimeScope.MostFutureContext.PatchManager.CreatePatch(ensemble);
			if (exception != null)
			{
				var unwrapedException = UnwrapException(exception);
				_dialogManager
					.CreateMessageDialog(
						unwrapedException.Message,
						"Could not compile or run the patch",
						DialogMode.Ok)
					.Show();

				_logger.Value.Log(exception.ToString());

				_runtimeScope.DiscardFutureContext();
				return;
			}

			// Wenn alles ok ist: Das Patch erstellen
			_runtimeScope.PrepareFutureContext();
			_runtimeScope.MostFutureContext.PatchManager.CreatePatch(
				ensemble,
				() =>
				{
					_runtimeScope.ApplyFutureContext();
					_logger.Value.Log("Ensemble created: " + ensemble.Name);
					SendMessage(new EnsembleRunningMessage(_ensembleManager.ActiveEnsemble));
				},
				false);
			SendMessage(new PreparePatchControlRequest());
		}

		private static Exception UnwrapException(Exception originalException)
		{
			var ex = originalException;
			while (true)
			{
				if (ex is DependencyResolutionException)
					ex = ex.InnerException;
				else if (ex is TargetInvocationException)
					ex = ex.InnerException;
				else
					return ex;
			}
		}
	}
}
