﻿// Copyright 2013 Ronald Schlenker, www.technewlogic.de.
// 
// This file is part of Byond.
// 
// GraphIT is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 2 of the License, or
// (at your option) any later version.
// 
// Foobar is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
using System;
using System.Reflection;
using System.Windows.Controls;
using System.Windows.Threading;
using Autofac;
using Byond.Recording;
using Byond.Shell.Views;
using Mvvm;
using Mvvm.Bindings;
using Byond.AudioMidi;
using Byond.EnsembleLibrary;
using Byond.EnsembleLibrary.ViewModels;
using Byond.Logging;
using Byond.OnlineDebugging.ViewModels;
using Byond.Persistence;
using Byond.HostObjects;
using Byond.Shell;
using Byond.Shell.Controls;
using Byond.Shell.ViewModels;
using Byond.UIRuntime;
using Utils.CtorSubstitution;
using Utils.DialogManagement;
using Utils.Messaging;

namespace Byond
{
	public class FrameworkBootstrapper : IDisposable
	{
		/// <param name="additionalRegistrations">
		/// IAudio and IMidi must be registered here.
		/// </param>
		public FrameworkBootstrapper(
			Action<ContainerBuilder> additionalRegistrations,
			Action<MenuItemRegistrar, ViewFactory> registerMenuItems)
		{
			var containerBuilder = new ContainerBuilder();
			containerBuilder.RegisterModule(new MessageInterceptionModule());
			containerBuilder.RegisterModule(new CtorSubstitutionModule());
			
			containerBuilder
				.RegisterAssemblyTypes(new[] { GetType().Assembly })
				.InstancePerDependency();
			containerBuilder.RegisterViewBindings(
				new SimplePropertyToViewBinding<AvalonEditor>(AvalonEditor.TextProperty));

			containerBuilder.RegisterType<Logger>().As<ILogger>().SingleInstance();
			containerBuilder.RegisterType<Ensemble>().FindConstructorsWith(BindingFlags.NonPublic);
			containerBuilder.RegisterType<DialogManager>().As<IDialogManager>().SingleInstance();
			containerBuilder.RegisterType<ByondControl>().FindConstructorsWith(BindingFlags.NonPublic).AsSelf().As<ContentControl>().SingleInstance();
			containerBuilder.Register(c => c.Resolve<ByondControl>().Dispatcher);
			containerBuilder.RegisterType<ObjectInspectorViewModel>().SingleInstance();
			containerBuilder.RegisterType<EnsembleManager>().AsSelf().As<IEnsembleCollection>().SingleInstance();
			containerBuilder.RegisterType<Library>().AsSelf().SingleInstance();
			containerBuilder.RegisterType<EnsembleUnitOfWork>().As<IEnsembleUnitOfWork>().SingleInstance();
			containerBuilder.RegisterType<Scheduler>().AsSelf().SingleInstance();
			containerBuilder.RegisterType<PersistenceManager>().ExternallyOwned().SingleInstance();
			containerBuilder.RegisterType<PatchManager>().FindConstructorsWith(BindingFlags.NonPublic).InstancePerLifetimeScope();
			containerBuilder.RegisterType<UiManager>().InstancePerLifetimeScope();
			containerBuilder.RegisterType<VoiceManager>().InstancePerLifetimeScope();
			containerBuilder.RegisterType<PatchInfo>().InstancePerLifetimeScope();
			containerBuilder.RegisterType<KeyboardCallbackManager>().AsSelf().InstancePerLifetimeScope();
			containerBuilder.RegisterType<MidiCallbackManager>().AsSelf().InstancePerLifetimeScope();
			containerBuilder.RegisterType<RuntimeScope>().SingleInstance();
			containerBuilder.RegisterType<RuntimeScopeContext>().InstancePerLifetimeScope();
			containerBuilder.RegisterType<EnsembleRunner>().ExternallyOwned().SingleInstance();
			containerBuilder.RegisterGeneric(typeof(CallbackController<>)).AsSelf().SingleInstance();
			
			containerBuilder.RegisterType<CodeEditorViewModel>().SingleInstance();
			containerBuilder.RegisterType<CodeEditorView>().SingleInstance();
			
			containerBuilder.RegisterType<LogViewModel>().SingleInstance();
			containerBuilder.RegisterType<LogView>().SingleInstance();
			
			containerBuilder.RegisterType<LiveViewModel>().InstancePerLifetimeScope();
			containerBuilder.RegisterType<LiveView>().InstancePerLifetimeScope();
			
			containerBuilder.RegisterType<BrowserViewModel>().AsSelf().SingleInstance();
			containerBuilder.RegisterType<DiagnosisManager>().AsSelf().SingleInstance();
			containerBuilder.RegisterType<MenuItemRegistrar>().AsSelf().SingleInstance();
			containerBuilder.RegisterType<PatchBuilder>().SingleInstance();
			containerBuilder.RegisterType<AudioRecorder>().SingleInstance();
			containerBuilder.RegisterType<Env>().SingleInstance();
			containerBuilder.RegisterType<SettingsManager>().SingleInstance();
			containerBuilder.Register(c => Container);

			additionalRegistrations(containerBuilder);

			Container = containerBuilder.Build();

			Container.Resolve<LogViewModel>();

			registerMenuItems(
				Container.Resolve<MenuItemRegistrar>(),
				Container.Resolve<ViewFactory>());

			Container.Resolve<PersistenceManager>();
			Container.Resolve<EnsembleRunner>();
			Container.Resolve<ByondControl>();
		}

		// TODO: Das ist nicht gut und wird nur gemacht, damit der NAudioPlayer resolved werden kann.
		public IContainer Container { get; private set; }

		public void Dispose()
		{
			if (Container != null)
				Container.Dispose();
		}
	}
}
