﻿// Copyright 2013 Ronald Schlenker, www.technewlogic.de.
// 
// This file is part of Byond.
// 
// GraphIT is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 2 of the License, or
// (at your option) any later version.
// 
// Foobar is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Interop;
using Autofac.Core;
using Byond.Compilation;
using Byond.EnsembleLibrary;
using Byond.HostObjects;
using Byond.Logging;
using Byond.PatchUiControls;
using Byond.Persistence;
using Byond.Shell.ViewModels;
using Byond.UIRuntime;
using Utils.DialogManagement;
using Utils.Messaging;
using Binder = Byond.UIRuntime.Binder;

namespace Byond
{
	sealed class PatchManager
	{
		private readonly IEnsembleCollection _ensembleCollection;
		private readonly Library _library;
		private readonly UiManager _uiManager;
		private readonly Lazy<VoiceManager> _voiceManager;
		private readonly Lazy<ILogger> _logger;
		private readonly PatchCompiler _currentPatchCompiler;
		private readonly PersistenceManager _persistenceManager;

		internal PatchManager()
		{
			RootPatch = Patch.Empty;
			RootPatchControl = new Control();
		}

		public event MessageEventHandler SendMessage;

		public Patch RootPatch { get; private set; }
		public Control RootPatchControl { get; private set; }

		private readonly List<MetaPatch> _patches = new List<MetaPatch>();
		internal IEnumerable<MetaPatch> Patches { get { return _patches; } }

		private Action _creationFinishedCallback;
		private bool _checkOnly;
		private Exception _rootPatchLoadedException;

		public Exception CreatePatch(Ensemble ensemble, Action creationFinishedCallback = null, bool checkOnly = true)
		{
			_checkOnly = checkOnly;

			try
			{
				// already initialized?
				if (_creationFinishedCallback != null)
					throw new Exception("PatchManager creation cycle already started!");
				_creationFinishedCallback = creationFinishedCallback ?? (() => { });

				var patch = _currentPatchCompiler.GeneratePatch(ensemble, _ensembleCollection.Ensembles);
				var metaPatch = _patches.First(it => it.Patch == patch);
				var patchControl = CreatePatchControl(metaPatch);

				// initialize the patch panels
				var i = 0;
				SetupPatchPanels(metaPatch, ref i);

				RootPatch = patch;
				RootPatchControl = patchControl;

				if (checkOnly)
				{
					using (var ps = new HwndSource(new HwndSourceParameters()))
					{
						// Force Load event
						RootPatchControl.Loaded += RootPatchControlOnLoaded;
						ps.RootVisual = new PresentationAnchor(RootPatchControl);
					}
				}
				else
				{
					RootPatchControl.Loaded += RootPatchControlOnLoaded;
				}

				return _rootPatchLoadedException;
			}
			catch (Exception e)
			{
				if (checkOnly)
					return e;
				throw;
			}
		}

		internal void RegisterPatch(Patch patch, Patch parent, string memberName)
		{
			var parentMetaPatch = (parent != null)
				? _patches.First(it => it.Patch == parent)
				: null;
			var metaPatch = new MetaPatch(patch, memberName, parentMetaPatch);
			_patches.Add(metaPatch);
		}

		private void SetupPatchPanels(MetaPatch metaPatch, ref int i)
		{
			i++;
			if (i > 100)
				throw new Exception("Maximum patch panel nesting reached (circular patch panels?).");

			var patchPanels = metaPatch.PatchControl
				.GetLogicalDescendants()
				.OfType<PatchPanel>()
				.Where(it => it.PatchControl == null)
				.ToList();
			foreach (var it in patchPanels)
			{
				var childPatch = metaPatch.FindDescendantByPath<MetaPatch>(it.Path);
				childPatch.PatchControl = CreatePatchControl(childPatch);
				it.PatchControl = childPatch.PatchControl;

				SetupPatchPanels(childPatch, ref i);
			}
		}

		private void RootPatchControlOnLoaded(object sender, RoutedEventArgs routedEventArgs)
		{
			// Komisch: Wenn hier eine Exception kommt, kann diese nicht oben (presentationSource)
			// abgefangen werden - die Anwendung schließt einfach.
			// deshalb: hier try-catch.

			try
			{
				RootPatchControl.Loaded -= RootPatchControlOnLoaded;
				SetupBindings();

				if (!_checkOnly)
				{
					Restore();
					_creationFinishedCallback();
				}
			}
			catch (Exception e)
			{
				_rootPatchLoadedException = e;
			}
		}

		private Control CreatePatchControl(MetaPatch metaPatch)
		{
			var ensemble = _ensembleCollection.Ensembles.FirstOrDefault(it =>
				it.FileName.GetFullTypeName(_library.LibraryDirectory)
					== metaPatch.Patch.GetType().FullName);

			var patchControl = !string.IsNullOrWhiteSpace(ensemble.UiCode)
				? PatchCompiler.GenerateUi(ensemble.UiCode)
				: new Control();
			metaPatch.PatchControl = patchControl;
			return patchControl;
		}

		private void SetupBindings()
		{
			// hook up the bindings
			var potentialBindings = UiBindingMarkupExtension
				.GetRegistrations()
				.Select(it => new
					{
						MarkupExtension = it,
						Patch = _patches.FirstOrDefault(it2 =>
							it2.PatchControl.GetLogicalDescendants(true).Contains(it.Parent))
					})
				.Where(it => it.Patch != null)
				.ToList();
			var bindingFields = new List<FieldAndPropWrapper>();
			foreach (var it in potentialBindings)
			{
				// create the binder...
				var codeBehindProperty = GetCodeBehindBindingTarget(it.MarkupExtension.BindingPath, it.Patch);
				var uiControlBindingTarget = new UiControlBindingTarget(it.MarkupExtension.Parent, it.MarkupExtension.Property);
				
				int? index;
				string s;
				it.MarkupExtension.BindingPath.TryGetBindingIndex(out index, out s);
				var bindingProperty = new Binder(uiControlBindingTarget, codeBehindProperty, index);
				_uiManager.RegisterBinding(bindingProperty);

				// which binding targets do we have?
				bindingFields.AddRange(codeBehindProperty.Targets);
			}

			// mark the property as bound in the meta patch
			_patches.ForEach(it => it.RegisterBoundProperties(bindingFields));

			// finally, remove the registrations
			UiBindingMarkupExtension.RemoveRegistration(
				potentialBindings.Select(it => it.MarkupExtension).ToList());
		}

		private static CodeBehindBindingTarget GetCodeBehindBindingTarget(string bindingPath, MetaPatch patch)
		{
			// TODO: Evtl. Multi-Binding machen?

			return new CodeBehindBindingTarget(
				bindingPath
					.Split(new[] { ";" }, StringSplitOptions.RemoveEmptyEntries)
					.Select(it =>
						{
							var fieldAndPropWrapper = patch.FindDescendantByPath<FieldAndPropWrapper>(it);
							var couldCreateBinding = fieldAndPropWrapper != null;
							return new
								{
									// TODO: Logging, wenn couldCreateBinding == false
									CouldCreateBinding = couldCreateBinding,
									Binding = fieldAndPropWrapper
								};
						})
					.Where(it => it.CouldCreateBinding)
					.Select(it => it.Binding)
					.ToList());
		}

		private void Restore()
		{
			if (_persistenceManager.PendingRestoreInfo == null)
			{
				_voiceManager.Value.EnableVoiceProcessing = true;
				return;
			}

			var patchDtos = _persistenceManager.PendingRestoreInfo;
			DoRestore(patchDtos);
			_persistenceManager.ClearPendingRestore();
			_voiceManager.Value.EnableVoiceProcessing = true;
		}

		private void DoRestore(IEnumerable<PatchDto> patchDtos)
		{
			var propsToMap =
				from dto in patchDtos
				let metaPatch = _patches.FirstOrDefault(it => it.RootPath == dto.RootPath)
				where metaPatch != null
				from metaProp in metaPatch.PersistProperties
				let dtoProp = dto.Properties.FirstOrDefault(it => it.Name == metaProp.MemberName)
				where dtoProp != null
				select new
					{
						metaProp,
						dtoProp
					};
			foreach (var it in propsToMap)
			{
				try
				{
					it.metaProp.SetValueMethod(it.dtoProp.Value, null);
				}
				catch (Exception e)
				{
					_logger.Value.Log("ERROR: Could not restore property: " + it.dtoProp.Name + " / " +
						it.dtoProp.Value + ": " + e.Message);
				}
			}
		}
	}
}