// Copyright 2013 Ronald Schlenker, www.technewlogic.de.
// 
// This file is part of Byond.
// 
// GraphIT is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 2 of the License, or
// (at your option) any later version.
// 
// Foobar is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
using System.Collections.Generic;
using System.Linq;
using Byond.HostObjects;
using Byond.Messages;
using Newtonsoft.Json;
using Utils;
using Utils.Messaging;

namespace Byond.Persistence
{
	public sealed class PersistenceManager
	{
		private readonly EnsembleManager _ensembleManager;
		private readonly RuntimeScope _runtimeScope;
		private readonly JsonSerializerSettings _jsonSerializerSettings;

		public PersistenceManager()
		{
			_jsonSerializerSettings = new JsonSerializerSettings
				{
					TypeNameHandling = TypeNameHandling.All
				};
		}

		public event MessageEventHandler SendMessage;

		public IEnumerable<PatchDto> PendingRestoreInfo { get; private set; }

		public bool ShouldPersistEnsemble
		{
			get { return _ensembleManager.ActiveEnsemble != null; }
		}

		public string PersistEnsembleToJson()
		{
			var dto = new EnsembleDto
				{
					EnsembleFileName = _ensembleManager.ActiveEnsemble.FileName,
					WasRunning = _runtimeScope.CurrentContext.PatchManager.RootPatch != Patch.Empty,
					Patches = PersistPatches().ToArray()
				};

			var json = JsonConvert.SerializeObject(dto, _jsonSerializerSettings);
			return json;
		}

		private IEnumerable<PatchDto> PersistPatches()
		{
			return _runtimeScope.CurrentContext.PatchManager.Patches
				.Select(it => it.ToDto())
				.ToList();
		}

		public void RestoreEnsemble(string json)
		{
			var dto = JsonConvert.DeserializeObject<EnsembleDto>(json, _jsonSerializerSettings);
			var ensemble = _ensembleManager.Ensembles
				.FirstOrDefault(it => it.FileName == dto.EnsembleFileName);
			if (ensemble == null)
				return;

			PendingRestoreInfo = dto.Patches;

			SendMessage(new ActivateEnsembleRequest(ensemble));
			if (dto.WasRunning)
				SendMessage(new RunActiveEnsembleRequest());
		}

		public void MarkForUpdate()
		{
			// TODO: Ex, wenn kein active ens
			PendingRestoreInfo = PersistPatches().ToList();
		}

		public void ClearPendingRestore()
		{
			PendingRestoreInfo = null;
		}
	}
}