// Copyright 2013 Ronald Schlenker, www.technewlogic.de.
// 
// This file is part of Byond.
// 
// GraphIT is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 2 of the License, or
// (at your option) any later version.
// 
// Foobar is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.IO;
using System.Linq;
using System.Windows;
using System.Windows.Input;
using Byond.EnsembleLibrary.ViewModels;
using Byond.HostObjects;
using Byond.Logging;
using Byond.Messages;
using Byond.OnlineDebugging.ViewModels;
using Byond.Persistence;
using Microsoft.Win32;
using Mvvm;
using Utils;
using Utils.DialogManagement;
using Utils.Messaging;

namespace Byond.Shell.ViewModels
{
	internal class CodeEditorViewModel : ViewModelBase
	{
		public CodeEditorViewModel(
			EnsembleManager ensembleManager,
			BrowserViewModel ensembleBrowser,
			DebugEditorViewModel debugEditor,
			IEnsembleUnitOfWork uow,
			MenuItemRegistrar menuItemRegistrar,
			Func<Ensemble, EnsembleEditorViewModel> ensembleEditorFactory,
			Env env)
		{
			_ensembleManager = ensembleManager;
			_ensembleEditorFactory = ensembleEditorFactory;
			EnsembleBrowser = ensembleBrowser;
			DebugEditor = debugEditor;

			_uow = uow;
			_uow.Changed += InvokePropertiesChanged;

			EnsembleBrowser.PropertyChanged += (s, e) => InvokePropertiesChanged();

			// Wenn das nicht gemacht wird, wird das DataTemplate im ItemsControl ignoriert.
			MenuItems = menuItemRegistrar.MenuItems
				.Select(it => new { UiElement = it() })
				.ToList();

			_openSessionDialog = new OpenFileDialog
				{
					InitialDirectory = env.PatchPath,
					Multiselect = false,
					DefaultExt = ".session",
					Filter = "Byond Sessions (.session)|*.session"
				};

			_saveSessionDialog = new SaveFileDialog
				{
					InitialDirectory = env.PatchPath,
					DefaultExt = ".session",
					Filter = "Byond Sessions (.session)|*.session"
				};
		}

		private readonly EnsembleManager _ensembleManager;
		private readonly Func<Ensemble, EnsembleEditorViewModel> _ensembleEditorFactory;
		private readonly IEnsembleUnitOfWork _uow;
		private readonly RuntimeScope _runtimeScope;
		private readonly PersistenceManager _persistenceManager;
		private readonly OpenFileDialog _openSessionDialog;
		private readonly SaveFileDialog _saveSessionDialog;
		private readonly IDialogManager _dialogManager;

		public event MessageEventHandler SendMessage;

		public BrowserViewModel EnsembleBrowser { get; private set; }

		private EnsembleEditorViewModel _ensembleEditor;
		public EnsembleEditorViewModel EnsembleEditor
		{
			get { return _ensembleEditor; }
			private set
			{
				_ensembleEditor = value;
				OnPropertyChanged("EnsembleEditor");
				OnPropertyChanged("CanGotoCode");
			}
		}

		public DebugEditorViewModel DebugEditor { get; private set; }

		public Visibility EnsembleEditorVisibility { get; private set; }
		public Visibility LiveEditorVisibility { get; private set; }
		public Visibility LogVisibility { get; private set; }

		public bool IsInCodeMode { get { return EnsembleEditorVisibility == Visibility.Visible; } }
		public bool IsInLiveMode { get { return LiveEditorVisibility == Visibility.Visible; } }
		public bool IsInDebugMode { get { return LogVisibility == Visibility.Visible; } }

		public IEnumerable<object> MenuItems { get; private set; }

		public void HandleMessage(EnsembleShownMessage message)
		{
			ShowEditor(_ensembleManager.EditingEnsemble);
			InvokePropertiesChanged();
		}

		public void HandleMessage(EnsemblesUpdatedMessage message)
		{
			InvokePropertiesChanged();
		}

		public void HandleMessage(EnsembleActivatedMessage message)
		{
			InvokePropertiesChanged();
		}

		public void HandleMessage(EnsembleStoppedMessage message)
		{
			InvokePropertiesChanged();
		}

		public void HandleMessage(EnsembleRunningMessage message)
		{
			InvokePropertiesChanged();
		}

		public void Run()
		{
			SendMessage(new RunActiveEnsembleRequest());
		}

		public void Stop()
		{
			SendMessage(new StopEnsembleRequest());
		}

		public void GotoLive()
		{
			SendMessage(new ShowViewRequest(_runtimeScope.CurrentContext.LiveView));
		}

		public void GotoLog()
		{
			SendMessage(new ShowViewRequest(typeof(LogViewModel)));
		}

		public void OpenSession()
		{
			if (!_openSessionDialog.ShowDialog().Value)
				return;
			_persistenceManager.RestoreEnsemble(File.ReadAllText(_openSessionDialog.FileName));
		}

		public void SaveSession()
		{
			if (_runtimeScope.CurrentContext.PatchManager.RootPatch == Patch.Empty)
			{
				_dialogManager
					.CreateMessageDialog(
						"Saving of sessions if only possible for a running patch.",
						"Cannot save session",
						DialogMode.Ok)
					.Show();
				return;
			}

			if (!_saveSessionDialog.ShowDialog().Value)
				return;
			if (string.IsNullOrWhiteSpace(_saveSessionDialog.FileName))
				return;

			var json = _persistenceManager.PersistEnsembleToJson();
			File.WriteAllText(_saveSessionDialog.FileName, json);
		}

		public bool CanRun { get { return _ensembleManager.ActiveEnsemble != null; } }
		public bool CanStop { get { return _runtimeScope.CurrentContext.PatchManager.RootPatch != Patch.Empty; } }
		public bool CanGotoLive { get { return CanStop; } }
		public bool CanSave { get { return _uow.DirtyEnsembles.Any(); } }
		public bool CanSaveAs { get { return _ensembleManager.EditingEnsemble != null; } }
		public bool CanSaveSession { get { return CanStop; } }

		private void ShowEditor(Ensemble ensemble)
		{
			EnsembleEditor = ensemble == null
				? null
				: _ensembleEditorFactory(ensemble);
		}

		private void InvokePropertiesChanged()
		{
			OnPropertyChanged("CanRun");
			OnPropertyChanged("CanStop");
			OnPropertyChanged("CanGotoLive");
			OnPropertyChanged("CanNew");
			OnPropertyChanged("CanSave");
			OnPropertyChanged("CanSaveAs");
		}
	}
}