// Copyright 2013 Ronald Schlenker, www.technewlogic.de.
// 
// This file is part of Byond.
// 
// GraphIT is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 2 of the License, or
// (at your option) any later version.
// 
// Foobar is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
using System;
using System.Reflection;
using Byond.PatchUiControls;
using Fasterflect;

namespace Byond.UIRuntime
{
	public sealed class Binder
	{
		public Binder(
			UiControlBindingTarget uiControlBindingTarget,
			CodeBehindBindingTarget codeBehindBindingTarget,
			int? index)
		{
			_parent = uiControlBindingTarget.Parent;
			_property = uiControlBindingTarget.Property;
			_codeBehindBindingTarget = codeBehindBindingTarget;
			_index = index;

			if (uiControlBindingTarget.Property.PropertyType == typeof(double))
				_smoother = new Smoother();

			if (_codeBehindBindingTarget.IsAction)
			{
				if (_parent is Button)
				{
					var button = _parent as Button;
					button.PreviewMouseLeftButtonDown += (s, e) =>
						{
							lock (this)
								_invokeActionOnNextBind = true;
						};
				}
				else
					throw new Exception("Action bindings are only supported on buttons.");
			}
		}

		private readonly UiControl _parent;
		private readonly PropertyInfo _property;
		private readonly CodeBehindBindingTarget _codeBehindBindingTarget;
		private readonly int? _index;
		private readonly Smoother _smoother;

		private bool _firstBind = true;
		private object _lastBindToPropertyValue;
		private object _lastCodeBehindPropertyValue;
		private bool _invokeActionOnNextBind;

		public void InvokeCLickAction()
		{
			if (!_invokeActionOnNextBind)
				return;

			_codeBehindBindingTarget.Invoke();
			lock (this)
				_invokeActionOnNextBind = false;
		}

		public void Smooth()
		{
			if (_smoother != null)
				_smoother.Process();
		}

		public void Bind()
		{
			var bindFromControlToCode = BindFromControlToCode();
			var bindFromCodeToControl = BindFromCodeToControl();

			// BindFromControlToCode (beim richtigen Property Change) hat immer Vorrang
			if (bindFromControlToCode.ShouldBind)
			{
				_codeBehindBindingTarget.SetValue(bindFromControlToCode.Value, _index);
				_lastCodeBehindPropertyValue = bindFromControlToCode.Value;
			}
			else if (bindFromCodeToControl.ShouldBind)
			{
				if (_smoother != null)
				{
					_smoother.ForceCurrentValue(
						(double)Convert.ChangeType(bindFromCodeToControl.Value, typeof(double)));
				}

				SetUiControlValue(bindFromCodeToControl.Value);
			}
		}

		private BindInfo BindFromControlToCode()
		{
			if ((_codeBehindBindingTarget == null || !_parent.WasChangedByUser)
				&& !_codeBehindBindingTarget.IsAction
				&& (_smoother != null && !_smoother.IsRunning))
			{
				return BindInfo.DontBind();
			}

			if (!_property.PropertyType.IsValueType && _property.PropertyType != typeof(string))
			{
				return BindInfo.DontBind();
			}

			if (_codeBehindBindingTarget.IsAction)
				return BindInfo.DontBind();

			if (_firstBind && _codeBehindBindingTarget.CanGetValue)
			{
				// After running the ensemble, we set the UI control
				// with the initial value of the patch.
				var initialValue = _codeBehindBindingTarget.GetValue(_index);
				// Das hier habe ich auskommentiert, denn es gab folgendes Problem:
				// Ein Button mit IsChecked Binding. Man musste nach dem ffnen 2 Mal
				// draufklicken, dass was pssiefrt ist.
				//SetUiControlValue(initialValue);
				if (_smoother != null && _parent.IsSmoothEnabled)
				{
					// kann auch float sein???
					_smoother.ForceCurrentValue(
						(double)Convert.ChangeType(initialValue, typeof(double)));
				}
			}
			_firstBind = false;

			object setValue;
			var propertyValue = _property.Get(_parent);
			if (_smoother != null && _parent.IsSmoothEnabled)
			{
				_smoother.SetDesiredValue((double)propertyValue);
				setValue = _smoother.CurrentValue;
			}
			else
				setValue = propertyValue;

			BindInfo result;
			if (!Equals(_lastBindToPropertyValue, propertyValue)
				|| (_smoother != null && (_smoother.IsRunning || _smoother.LastStepFlag)))
			{
				result = BindInfo.PropertyValueChanged(setValue);
				if (_smoother != null)
					_smoother.LastStepFlag = false;
			}
			else
			{
				result = BindInfo.DontBind();
			}
			_lastBindToPropertyValue = propertyValue;

			// quittieren
			_parent.WasChangedByUser = false;

			return result;
		}

		private BindInfo BindFromCodeToControl()
		{
			if (_codeBehindBindingTarget == null)
				return BindInfo.DontBind();
			if (!_codeBehindBindingTarget.CanGetValue)
				return BindInfo.DontBind();
			if (!_parent.CanBindFrom)
				return BindInfo.DontBind();

			var propertyValue = _codeBehindBindingTarget.GetValue(_index);
			if (Equals(_lastCodeBehindPropertyValue, propertyValue))
				return BindInfo.DontBind();
			_lastCodeBehindPropertyValue = propertyValue;
			return BindInfo.PropertyValueChanged(propertyValue);
		}

		private void SetUiControlValue(object value)
		{
			if (!_property.PropertyType.IsInstanceOfType(value))
				value = Convert.ChangeType(value, _property.PropertyType);

			_property.Set(_parent, value);
		}
	}
}