// Copyright 2013 Ronald Schlenker, www.technewlogic.de.
// 
// This file is part of Byond.
// 
// GraphIT is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 2 of the License, or
// (at your option) any later version.
// 
// Foobar is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
using System;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using Autofac;
using Utils;

namespace Mvvm
{
	public class ViewFactory
	{
		internal ViewFactory(
			ILifetimeScope lifetimeScope,
			ViewBindingConfiguration viewBindingConfiguration)
		{
			_lifetimeScope = lifetimeScope;
			_viewBindingConfiguration = viewBindingConfiguration;
		}

		private readonly ILifetimeScope _lifetimeScope;
		private readonly ViewBindingConfiguration _viewBindingConfiguration;

		public object GetViewForModel<TViewModel>()
		{
			var viewModelPropertyType = typeof(TViewModel);
			return GetViewForModel(viewModelPropertyType);
		}

		// TODO: Generalisieren mit GetViewForModel(Type viewModelPropertyType)
		public object GetViewForModel(object viewModel)
		{
			// TODO: viewModel == null??

			var viewModelPropertyType = viewModel.GetType();
			return DoGetViewForModel(viewModel, viewModelPropertyType);
		}

		public object GetViewForModel(Type viewModelPropertyType)
		{
			var viewModel = _lifetimeScope.Resolve(viewModelPropertyType);
			return DoGetViewForModel(viewModel, viewModelPropertyType);
		}

		private object DoGetViewForModel(object viewModel, Type viewModelPropertyType)
		{
			// TODO: PropertyChanged
			// TODO: CollectionChanged

			// TODO: Soll schon vom MvvmBootstrapper gemacht werden

			var viewTypeName = GetViewTypeName(viewModelPropertyType);
			var viewType = viewModelPropertyType.Assembly.GetTypes()
				.FirstOrDefault(it => it.FullName == viewTypeName);
			// TODO: throw-continue?
			if (viewType == null)
				throw new Exception("View '" + viewTypeName + "' was not found.");

			// TODO: Nicht einfach casten, sondern schauen bzw. nochmal berlegen, ob einfach weitergemacht werden soll
			var view = (viewType.IsPrimitive || viewType == typeof(string))
				? new TextBlock { Text = viewModel.ToString() }
				: _lifetimeScope.Resolve(viewType);
			// IMP: Sehr hnlich die Methode opben: throw-continue?
			if (view is FrameworkElement)
			{
				var frameworkElementView = (FrameworkElement)view;

				// Ganz wichtig, weil das Loaded Event mehrmals kommen kann
				RoutedEventHandler onLoaded = null;
				onLoaded = (s, e) =>
					{
						frameworkElementView.Loaded -= onLoaded;
						SetupView(viewModel, frameworkElementView);

						if (view is IMvvmView)
							((IMvvmView)view).OnViewSetupComplete();
					};
				frameworkElementView.Loaded += onLoaded;
			}

			return view;
		}

		private static string GetViewTypeName(Type viewModelType)
		{
			var namespaceName = viewModelType.Namespace.EndsWith("ViewModels")
				? viewModelType.Namespace.Substring(0, viewModelType.Namespace.Length - 6) + "s"
				: viewModelType.Namespace;
			var typeName = viewModelType.Name.EndsWith("ViewModel")
				? viewModelType.Name.Substring(0, viewModelType.Name.Length - 5)
				: viewModelType.Name;

			return namespaceName + "." + typeName;
		}

		private void SetupView(object viewModel, FrameworkElement view)
		{
			view.DataContext = viewModel;

			var viewType = view.GetType();

			// Call InitializeComponent
			var initializeComponentMethod = viewType.GetMethod("InitializeComponent", new Type[] { });
			if (initializeComponentMethod == null)
				return;
			initializeComponentMethod.Invoke(view, new object[] { });

			// All View Managers can setup their conventions of the view
			_viewBindingConfiguration.Bindings
				.ForEach(it => it.SetupView(this, view, viewModel));

			// TODO: Unloaded Event
		}
	}
}